/*
 * Copyright (c) 2001 Massachusetts Institute of Technology.
 *
 * This file is part of VideoCM. 
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `COPYING' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file AUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the AUTHORS file.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


/*
 *     rtsp_player.cc
 *     Nick Feamster (feamster@lcs.mit.edu)
 *     MIT Lab for Computer Science
 *     created July 2000
 */


#include "rtsp.h"
#include <tcl.h>
#include <tk.h>

const char  sContentLength[] = "Content-Length";
const char  sSession[] = "Session";
const char  sContentType[] = "Content-Type";

void tcpread_callback (ClientData cd, int mask) {
  RTSP_Player *rtsp = (RTSP_Player*)cd;
  rtsp->tcpread_cb(cd,mask);
}

void adu_callback(ClientData cd, int mask) {
  RTSP_Player *rtsp = (RTSP_Player*)cd;
  rtsp->read_adu();
}


RTSP_Player::RTSP_Player() {
  assert(offsetof(RTP_HDR, ssrc) == 8);

  authenticated = 0;
  udp_fd = 0;
  p_state.flags = 0;
  p_state.url[0] = '\0';
  p_state.serverurl[0] = '\0';
  p_state.cur_state = INIT_STATE;
  p_state.streams = 0;

  for (int i=0; i<ADU_BUFSIZE; i++) 
    adu_seqno[i]=-1;
  playout_start_ = 0;
  last_adu_ = -1;

  decode_fp_ = fopen("/tmp/mpeg_bitstream.bits", "wb");

#ifdef AUDIO
  wav_session = 0;
  wav_fd = 0;
  wav_rate = 0;
  wav_sampleformat = 0;
#endif

#ifdef _USE_CM_
  rrtp_chan = srrtp_create_cm_channel(CM_BUF_SOCK, CM_CC_AIMD);
#else
  rrtp_chan = srrtp_create_channel();
#endif

#ifdef TCL_GUI
  interface_init();
#endif

}


RTSP_Player::~RTSP_Player() {
  fclose(decode_fp_);
}




/******************************************
 *        housekeeping functions
 ******************************************/


void
RTSP_Player::assign_udpfd(int fd)
{
  assert(fd <= MAX_FDS);
  udp_fd = fd;

  // this is legacy stuff.  the old code used a callback waiting on the data socket
  // and made the callback at the application layer, but this is wrong since
  // this receive should be handled at the RRTP layer.
  
  // Tcl_CreateFileHandler(udp_fd, TCL_READABLE, (Tcl_FileProc*)udpread_callback, (ClientData)this);
}


void
RTSP_Player::udpread_callback(int seqno, int len, int layer) {
  //udp_data_recv(udp_fd);
  rtp_data_handle(seqno, len, layer);
}



void
RTSP_Player::tcpread_cb(ClientData cdata, int mask) {
  io_read(ssock_);
  msg_handler();
}



void
RTSP_Player::clear_udpfd(void)
{
  int i;
//    if (udp_fd)
//      udp_close(udp_fd);
  udp_fd = 0;
}


/************************************************************
 *   player specific support and message building routines
 ***********************************************************/

int
RTSP_Player::end_of_streams(STREAM * s)
{
  int             r = 1;

  /*
   * returns !0 if at end of play, 0 means not last packet for all streams in
   * the session.
   */

  /* checking if the server has sent last frame status for each stream */
  while (s) {
    if (!(s->stream_flags & STREAM_LAST_PACKET)) {
      r = 0;
      break;
    }
    s = s->next;
  }

  return (r);
}

int
base64_to_bin(u_char c)
{
  int             r;

  if ((c >= 'A') && (c <= 'Z'))
    r = (int) (c - 'A');
  else if ((c >= 'a') && (c <= 'z'))
    r = (int) (c - 'a') + 26;
  else if ((c >= '0') && (c <= '9'))
    r = (int) (c - '0') + 52;
  else if (c == '+')
    r = 62;
  else if (c == '/')
    r = 63;
  else
    r = -1;			/* not a base64 character */

  return (r);
}


int
decode_base64(u_char * o, char *i, int len)
{
  int             n;
  int             v;
  u_int32         a = 0;	/* accumulated value */
  int             s = 0;	/* shift register */

  for (n = 0; n < len; i++) {
    v = base64_to_bin(*i);
    if (v == -1)
      return (0);		/* hit an invalid base64 character */

    a <<= 6;			/* accumulate binary value */
    s += 6;
    a |= (u_int32) v;
    if (s >= 8) {
      s -= 8;
      v = a >> s;
      *o++ = (u_char) (v & 0xFFL);
      n++;
    }
  }

  return (1);
}


int
RTSP_Player::set_one_stream(char *b, int len, STREAM * s)
{
  char           *p, *n, *e;	/* buffer pointers */
  int             sl;		/* search limit */
  int             sdl;		/* stream descriptor length */
  PCM_HDR         pcm_header;	/* header description from opaque data */

  /*
   * return 0 if invalid or incomplete property settings are encountered.
   * Otherwise, return the length of the stream descriptor.
   */

  if (DescriptionFormat == DESCRIPTION_MH_FORMAT) {
    /* locate the beginning of the Media Stream Descriptor. */
    if ((p = get_string(b, len, "<MediaStream", " \t\r\n")) == 0) {
      printf("ALERT: Missing MediaStream description.\n");
      return (0);
    }
    sl = len - (int) (p - b);	/* adjust search length for new reference. */

    /* locate the end of the Media Stream descriptor */
    if ((e = get_string(p, sl, ">", " \t\r\n")) == 0) {
      printf("ALERT: Missing Media Stream description termination.\n");
      return (0);
    }
    sdl = (int) (e - b);	/* length of stream description */
    /* leave p pointing to start of stream descriptor.  We'll use it later. */

    /*
     * set stream id.  The player can multiplex streams on the data channel
     * using this ID.  The server will send the stream ID in the RTP packet
     * header on the data channel.
     */
    if (!get_int(p, sl, "StreamID", " =\t", (int *) &s->stream_id, 0)) {
      printf("ALERT: Missing StreamID from Media Stream description.\n");
      return (0);
    }
    /* set mime type */
    if ((n = get_string(p, sl, "MimeType", " =\t")) == 0) {
      printf("ALERT: Missing MimeType from Media Container.\n");
      return (0);
    }
    /* for now I only accept text or audio MIME types. */
    if (strncasecmp(n, "plain/text", 4) == 0)
      s->mime_type = MIME_TEXT;
    else if (strncasecmp(n, "audio/x-pn-wav", 14) == 0)
      s->mime_type = MIME_AUDIO;
    else {
      printf("ALERT: MimeType not supported.  Player supports only `Text' or "
	     "'audio/x-pn-wav'.\n");
      return (0);
    }

    if (!get_int(p, sl, "MaxBitRate", " =\t", 0, &s->max_bit_rate)) {
      printf("ALERT: MaxBitRate not found in <MediaStream> descriptor.");
      return (0);
    }
    if (!get_int(p, sl, "MaxPktSize", " =\t", 0, &s->max_pkt_size)) {
      printf("ALERT: MaxBitRate not found in <MediaStream> descriptor.");
      return (0);
    }
  } else {
    /* SDP format. */
    
    /* locate the beginning of the Media Stream Descriptor. */
    if ((p = get_string(b, len, "m=", "")) == 0) {
      printf("ALERT: Missing MediaStream description.\n");
      return (0);
    }
    p -= 2;
    sl = len - (int) (p - b);	/* adjust search length for new reference. */

    /* locate the end of the Media Stream descriptor */
    if ((e = get_string(p + 2, sl - 2, "m=", "")) == 0) {
      /* Ok, it's the last Media stream description. */
      e = p + sl - 1;
    } else
      e -= 2;

    sdl = (int) (e - b);	/* length of stream description */
    /* leave p pointing to start of stream descriptor.  We'll use it later. */

    /* set mime type */
    if ((n = get_string(p, sl, "m=", "")) == 0) {
      printf("ALERT: Missing MimeType from Media Container.\n");
      return (0);
    }
    /* for now I only accept text or audio MIME types. */
    if (strncasecmp(n, "other_media_type", 16) == 0)
      s->mime_type = MIME_TEXT;
    else if (strncasecmp(n, "audio 0 RTP/AVP", 15) == 0)
      s->mime_type = MIME_AUDIO;
    else if (strncasecmp(n, "video 0 RTP/AVP", 15) == 0)
      s->mime_type = MIME_VIDEO;
    else {
      printf("ALERT: MimeType not supported.\n");
      printf("Player supports only audio 0 RTP/AVP 0 or other_media_type.\n");
      return (0);
    }
    if (!get_int(p, sl, "audio 0 RTP/AVP", " \t",
		 (int *) &curr_payload_type, 0) &&
	!get_int(p, sl, "video 0 RTP/AVP", " \t",
		 (int *) &curr_payload_type, 0)) {
      printf("ALERT: Missing payload type from Media Stream description.\n");
      return (0);
    }
    switch (curr_payload_type) {
    case (RTP_PAYLOAD_PCMU):	/* u-law */
      s->max_bit_rate = 64000;
      s->max_pkt_size = 400;
      break;
    case (RTP_PAYLOAD_PCMA):	/* a-law */
      s->max_bit_rate = 64000;
      s->max_pkt_size = 400;
      break;
    case (RTP_PAYLOAD_L16_2):	/* linear 16, 44.1khz, 2 channel */
      s->max_bit_rate = 705600;
      s->max_pkt_size = 4410;
      break;
    case (RTP_PAYLOAD_L16_1):	/* linear 16, 44.1khz, 1 channel */
      s->max_bit_rate = 705600;
      s->max_pkt_size = 4410;
      break;
    case (RTP_PAYLOAD_RTSP):	/* 101, Dynamic. */
      if (!get_int(p, sl, "MaxBitRate", ":\t", 0, &s->max_bit_rate)) {
	printf("ALERT: MaxBitRate not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "MaxPktSize", ":\t", 0, &s->max_pkt_size)) {
	printf("ALERT: MaxPktSize not found in descriptor.\n");
	return (0);
      }
      break;
    case (RTP_PAYLOAD_MPEG4):	/* 102, Dynamic. */
      if (!get_int(p, sl, "Width", ":\t", 0, &s->frame_width)) {
	printf("ALERT: Display width not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "Height", ":\t", 0, &s->frame_height)) {
	printf("ALERT: Display height not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "Framerate", ":\t", 0, &s->framerate)) {
	printf("ALERT: Framerate not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "Duration", ":\t", 0, &s->duration)) {
	printf("ALERT: Duration not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "Bits-per-pixel", ":\t", 0, &s->bpp)) {
	printf("ALERT: Bits-per-pixel not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "VOs", ":\t", 0, &s->num_vos)) {
	printf("ALERT: Number of Video Objects not found in descriptor.\n");
	return (0);
      }
      if (!get_int(p, sl, "Post-Filter-Type", ":\t", 0, &s->post_filter_type)) {
	printf("ALERT: Post-Filter-Type not found in descriptor.\n");
	return (0);
      }



      char token[7];
      sprintf(token, "VOLs(");
      for (int i=0; i<s->num_vos; i++) {
	sprintf(token+5, "%d)", i);
	if (!get_int(p, sl, token, ":\t", 0, &s->num_vols[i])) {
	  printf("ALERT: Number of VOLs for Object #%dnot found in descriptor.\n", i);
	  return (0);
	}
//  	else {
//  	  printf("Number of VOLs for Object #%d --> %d.\n", i, s->num_vols[i]);
//  	}
      }

      frametick_ = (int)(((double)1/s->framerate)*1000 + 1);
      break;
    }
  }
 
  if (s->mime_type == MIME_AUDIO) {
	switch (curr_payload_type) {
	case (RTP_PAYLOAD_PCMU):	/* u-law */
		PCM_hdr.wFormatTag = 7;
		PCM_hdr.nChannels = 1;
		PCM_hdr.nSamplesPerSec = 8000;
		PCM_hdr.wBitsPerSample = 8;
		PCM_hdr.nBlockAlign = 1; 
		PCM_hdr.nAvgBytesPerSec = 8000;   
      break;
    case (RTP_PAYLOAD_PCMA):	/* a-law */
		PCM_hdr.wFormatTag = 6;
		PCM_hdr.nChannels = 1;
		PCM_hdr.nSamplesPerSec = 8000;
		PCM_hdr.wBitsPerSample = 8;
		PCM_hdr.nBlockAlign = 1; 
		PCM_hdr.nAvgBytesPerSec = 8000;    
      break;
    case (RTP_PAYLOAD_L16_2):	/* linear 16, 44.1khz, 2 channel */
		PCM_hdr.wFormatTag = 1;
		PCM_hdr.nChannels = 2;
		PCM_hdr.nSamplesPerSec = 44100;
		PCM_hdr.wBitsPerSample = 16;
		PCM_hdr.nBlockAlign = 4; 
		PCM_hdr.nAvgBytesPerSec = 176400;    
      break;
    case (RTP_PAYLOAD_L16_1):	/* linear 16, 44.1khz, 1 channel */
		PCM_hdr.wFormatTag = 1;
		PCM_hdr.nChannels = 1;
		PCM_hdr.nSamplesPerSec = 44100;
		PCM_hdr.wBitsPerSample = 16;
		PCM_hdr.nBlockAlign = 2; 
		PCM_hdr.nAvgBytesPerSec = 88200;    
	break;
    case (RTP_PAYLOAD_RTSP):	/* 101, Dynamic. */
			if (DescriptionFormat == DESCRIPTION_MH_FORMAT) {
			  if ((n = get_string(p, sl, "TypeSpecificData", " =\t")) == 0) {
			printf("ALERT: Missing TypeSpecificData needed to set audio "
				   "parameters.\n");
			return (0);
			  }
			} else {
			  if ((n = get_string(p, sl, "a=TypeSpecificData", ":\t")) == 0) {
			printf("ALERT: Missing TypeSpecificData needed to set audio "
				   "parameters.\n");
			return (0);
			  }
			}

			if (*n != '"') {
			  printf("ALERT: Missing starting quote character from TypeSpecificData"
				 " string.\n");
			  return (0);
			}
			/*
			 * determine length of quoted Base64 encoded string. The stream
			 * descriptor terminator ">" is still present
			 */
			n++;
			if (!(e = strpbrk(n, "=> \"\r\n\t"))) {
			  printf("ALERT: Missing terminator character to Base64 encoded "
				 "string.\n");
			  return (0);
			}
			/* decode the PCM header */
			assert((e - n) == (sizeof(PCM_HDR) * 4 / 3));
			if (!decode_base64((u_char *) & pcm_header, n, sizeof(PCM_HDR))) {
			  printf("ALERT: Encountered an invalid Base64 character in encoded "
				 "string.\n");
			  return (0);
			}
			/* the server will put the pcm_header in NBO order prior to encoding. */
			if (!big_endian()) {
			  swap_word(&pcm_header.usFormatTag, 1);
			  swap_word(&pcm_header.usChannels, 1);
			  swap_dword(&pcm_header.ulSamplesPerSec, 1);
			  swap_word(&pcm_header.usBitsPerSample, 1);
			  swap_word(&pcm_header.usSampleEndianness, 1);
			}
			assert(pcm_header.usChannels != 0);
			/* Setup WAVEFORMATEX structure to be used by CWaveOutput::ParseWavData */
			PCM_hdr.wFormatTag = pcm_header.usFormatTag;
			PCM_hdr.nChannels = pcm_header.usChannels;
			PCM_hdr.nSamplesPerSec = pcm_header.ulSamplesPerSec;
			PCM_hdr.wBitsPerSample = pcm_header.usBitsPerSample;

			assert(PCM_hdr.nChannels != 0);
			PCM_hdr.nBlockAlign = (PCM_hdr.nChannels * PCM_hdr.wBitsPerSample) / 8;
			PCM_hdr.nAvgBytesPerSec = PCM_hdr.nSamplesPerSec * PCM_hdr.nBlockAlign;
      break;
    }
	PCM_hdr.CbSize = 0;
	s->PCM_hdr.usFormatTag = PCM_hdr.wFormatTag;
	s->PCM_hdr.usChannels = PCM_hdr.nChannels;
	s->PCM_hdr.ulSamplesPerSec = PCM_hdr.nSamplesPerSec;
	s->PCM_hdr.usBitsPerSample = PCM_hdr.wBitsPerSample;
    s->PCM_hdr.usSampleEndianness = 1;
  }    
  return (sdl);
}

/* Here, we deal with media announcements of each stream. */
int
RTSP_Player::set_stream_settings(char *b, int len, struct SESSION_STATE * state)
{
  STREAM         *s;		/* stream descriptor pointer */
  char           *p, *pp;	/* buffer pointer */
  int             sl;		/* search limit */
  int             sdl;		/* stream descriptor length */
  int             i;
  int             num_streams;

  /* return 0 if invalid or incomplete settings. */

  if (DescriptionFormat == DESCRIPTION_MH_FORMAT) {
    if ((p = get_int(b, len, "Duration", " =\t", 0,
		     &state->settings.duration)) == 0) {
      printf("ALERT: Missing Duration from Media Container.\n");
      return (0);
    }
    if ((p = get_int(b, len, "NumStreams", " =\t", &num_streams, 0)) == 0) {
      printf("ALERT: Missing NumStreams from Media Container.\n");
      return (0);
    }
    sl = len - (int) (p - b);	/* adjust search length for new reference. */
  } else {
    num_streams = 0;
    for (pp = b, sl = len; (p = get_string(pp, sl, "m=", ""));) {
      sl -= (int) (p - pp);
      num_streams++;
      pp = p;
    }
  }

  if (num_streams == 0) {
    printf("ALERT: Missing NumStreams from Media Container.\n");
    return (0);
  }
  state->settings.num_streams = num_streams;

  /* locate the beginning of the MediaStreams description section. */
  if (DescriptionFormat == DESCRIPTION_MH_FORMAT) {
    if ((p = get_string(p, sl, "<MediaStreams>", " \t\r\n")) == 0) {
      printf("ALERT: Missing <MediaStreams> section description.\n");
      return (0);
    }
  } else {
    if ((p = get_string(b, len, "m=", "")) == 0) {
      printf("ALERT: Missing Media Announcement section description.\n");
      return (0);
    }
    p -= 2;
  }

  sl = len - (int) (p - b);	/* adjust search length for new reference. */

  for (i = 0; i < num_streams; i++, p += sdl) {
    /*
     * allocate stream descriptor.  Stream descriptors are freed by
     * free_streams().
     */
    s = (STREAM*)calloc(1, sizeof(STREAM));
    s->next = state->streams;
    state->streams = s;		/* add new descriptor to head of list. */

    /* collect & set one set of stream properties */
    sdl = set_one_stream(p, sl, s);
    sl -= sdl;
    if (!sdl || (sl < 0))
      return (0);
  }

  /* enough for now. */
  return (1);
}



/******************************************
 *         send message functions
 ******************************************/
void 
/* BEGIN Edited By Ed Hogan TIS Inc. */
RTSP_Player::send_hello_request( const char *sUrl)
{
   char b [256];
   if (sUrl == NULL) 
      sprintf( b, "%s * %s %d\n", HELLO_TKN, 
			RTSP_VER, RTSP_send_seq_num++ );
   else 
      sprintf( b, "%s %s %s %d\n", \
         HELLO_TKN, sUrl, RTSP_VER, RTSP_send_seq_num++ );
   strcat(b, "User-Agent: RTSP-Reference/0.4alpha ");
#ifdef FREEBSD
	strcat(b, "(FreeBSD)");	
#endif

#ifdef SOLARIS
	strcat(b, "(Solaris)");
#endif

#ifdef LINUX
	strcat(b, "(LINUX)");
#endif

#ifdef WIN32
	strcat(b, "(Win)");
#endif

  strcat(b, "\r\n\r\n");

   bwrite( b, (u_short) strlen( b ) );
      RTSP_last_request = RTSP_HELLO_METHOD;
}
/* END   Edited By Ed Hogan TIS Inc. */

void
RTSP_Player::send_close_request(void)
{
  char            b[256];

  sprintf(b, "%s * %s %d\r\n\r\n", CLOSE_TKN, RTSP_VER, RTSP_send_seq_num++);
  bwrite(b, (u_short) strlen(b));
  RTSP_last_request = RTSP_CLOSE_METHOD;
  authenticated = 0;
}

void
RTSP_Player::send_get_request(const char *sUrl)
{
  char            b[256];

  /* save the url string for future use in setup request. */
  strcpy(p_state.url, p_state.serverurl);
  strcat(p_state.url, sUrl);
  sprintf(b, "%s %s %s %d\n", GET_TKN, p_state.url, 
		RTSP_VER, RTSP_send_seq_num++);
  /* For Test only.
  strcat(b, "PEP: {{strength must}}\n");
  */
  /*
   * NT rmserver did not work with the Accept line in the request message.
   * The RTSP spec lists it as required on the request.  To include it remove
   * the comment mark.
   */
  strcat(b, "Accept: application/sdp; application/x-rtsp-mh\n");
  strcat(b, "\r\n\r\n");
  bwrite(b, (u_short) strlen(b));
  RTSP_last_request = RTSP_GET_METHOD;
}

/*
 * The final version should deal with multi-streams, i.e., send setup request
 * for each streams for the media.
 */
void
RTSP_Player::send_setup_request(void)
{
  char            b[256];

  sprintf(b, "%s %s %s %d\n", SETUP_TKN, p_state.url, RTSP_VER,
	  RTSP_send_seq_num++);

  /* make port assignment for data stream(s). */

#ifdef _USE_CM_
  //assign_udpfd(udp_open());
  // this does a re-binding, etc. so maybe should only be called once?
  srrtp_cm_listen(rrtp_chan, 5001, 5002, &udp_fd);
#else
  srrtp_listen(rrtp_chan, 5001, 5002, &udp_fd);
  //assign_udpfd(udp_open());
#endif

  p_state.settings.data_port = udp_getport(udp_fd, &p_state.settings.saddr_in);

  sprintf(b + strlen(b), "Transport: rtp/udp;port=%d\r\n\r\n",
	  p_state.settings.data_port);

  bwrite(b, (u_short) strlen(b));
  RTSP_last_request = RTSP_SETUP_METHOD;
}

void
RTSP_Player::send_play_range_request(char *range)
{
  char            b[256];

  sprintf(b, "%s %s %s %d\n", PLAY_TKN, p_state.url, RTSP_VER,
	  RTSP_send_seq_num++);
  /* Add the session#, this is a dummy version. */
  if (p_state.flags & SS_USE_SESSION_ID)
    add_session_id(b, &p_state);

  if (range && *range)
    sprintf(b + strlen(b), "range: %s\n", range);
  else if (p_state.flags & SS_PLAY_RANGE)
    add_play_range(b, &p_state);
  else if (end_of_streams(p_state.streams))	/* start from beginning
						 * again. */
    sprintf(b + strlen(b), "Range: smpte=00:00:00-\n");
  strcat(b, "\r\n\r\n");

  bwrite(b, (u_short) strlen(b));
  RTSP_last_request = RTSP_PLAY_METHOD;
}

void
RTSP_Player::send_pause_request(void)
{
  char            b[256];

  sprintf(b, "%s %s %s %d\r\n\r\n", PAUSE_TKN, p_state.url, RTSP_VER,
	  RTSP_send_seq_num++);
  bwrite(b, (u_short) strlen(b));
  RTSP_last_request = RTSP_PAUSE_METHOD;
}


/******************************************
 *   received message handler functions
 ******************************************/

 /* Parse the PEP: (including PEP-Info:, C-PEP: and C-PEP-Info:)
	header and see if they contain {strength must} which is not allowed. 
	return 0 if OK, -1 if encounters any, -2 if illegal PEP. */
int
RTSP_Player::handle_PEP(char *buf, int StrLen)
{
	int PEP_Start;
	int PEP_End = 0;
	int ParenCount = 1;
	char *TempStr;
	char cc;

	if ((TempStr = strstr(buf, "PEP:")) == NULL)
		if ((TempStr = strstr(buf, "PEP-Info:")) == NULL)
			if ((TempStr = strstr(buf, "C-PEP:")) == NULL)
				TempStr = strstr(buf, "C-PEP-Info:");
	if (TempStr == NULL) 
		return 0;
	PEP_Start = TempStr - buf;
	TempStr = strstr(buf + PEP_Start, "{");
	if (TempStr == NULL) 
		return -2;
	
	PEP_End = TempStr - buf;
	do {
		PEP_End++;
		if (PEP_End > StrLen)
			return -2;
		cc = *(++TempStr);
		ParenCount+= (cc == '{') - (cc == '}');
	} while (ParenCount);

	if (strstr(buf + PEP_Start, "{strength must}") == NULL)
		return 0;
	else 
		return -1;
}

void
RTSP_Player::handle_hello_request(void)
{
  interface_hello(OpenFileName);
  discard_msg();		/* remove remainder of message from in_buffer */
  /* here's were you can add authentication checks and rejection */

  send_reply(200, 0);		/* OK */
}

void
RTSP_Player::handle_hello_reply(int status)
{
  discard_msg();		/* remove remainder of message from in_buffer */

  if (status < 203)
    authenticated = 1;

  /*if (OpenFileName != NULL) 
	  send_get_request(OpenFileName);*/
}


int
RTSP_Player::handle_get_reply(int status)
{
  char           *p;
  char            b[64];
  int             n;
  int             hl;		/* message header length */
  int             bl;		/* message body length */

  /* return !0 if get response is dealt with and setup should be sent. */
  if (status > 202) {
    discard_msg();		/* remove remainder of message from in_buffer */
    return (0);
  }
  /* free up stream descriptors from any previous get. */
  free_streams(&p_state);

  get_msg_len(&hl, &bl);	/* set header and message body length */

  /* get the content type */
  if ((p = get_string(in_buffer, hl, (char *) sContentType, " :")) == 0) {
    printf("ALERT: required Content-Type header field is missing.\n");
    remove_msg(hl + bl);	/* remove remainder of message from in_buffer */
    return (0);
  }
  if ((n = sscanf(p, " %60s ", b)) != 1) {
    printf("ALERT: Content-Type value string not present\n");
    remove_msg(hl + bl);	/* remove remainder of message from in_buffer */
    return (0);
  }

   if (handle_PEP(in_buffer, hl) == -1) { 
    remove_msg(hl + bl);	/* remove remainder of message from in_buffer */
	send_reply(420, 0);
    return (0);
  }

  if (strcasecmp(b, "application/sdp")) {
    if (strcasecmp(b, "application/x-rtsp-mh")) {
      printf("ALERT: Content-Type (%s) not recognized.\n", b);
      remove_msg(hl + bl);
      /* remove remainder of message from in_buffer */
      return (0);
    }
    DescriptionFormat = DESCRIPTION_MH_FORMAT;
  } else
    DescriptionFormat = DESCRIPTION_SDP_FORMAT;

  /* get the media announcements for each stream. */
  if (!set_stream_settings(&in_buffer[hl], bl, &p_state)) {
    remove_msg(hl + bl);	/* remove remainder of message from in_buffer */
    return (0);
  }
  remove_msg(hl + bl);		/* remove remainder of message from in_buffer */

  return (1);
}

void
RTSP_Player::handle_setup_reply(int status)
{
  int             hl;		/* message header length */
  int             bl;		/* message body length */
  int             val;

  if (status < 203) {
    p_state.settings.session_id = 0;
    interface_new_session(0);
    get_msg_len(&hl, &bl);	/* set header and message body length */
    if (get_int(in_buffer, hl, (char *) sSession, " :", &val, 0)) {
      p_state.flags |= SS_USE_SESSION_ID;
      val = 0; /* This session is diffrent from the session in the reply. */
      /* Dummy version of session_id. */
      p_state.settings.session_id = val;
    } else
      p_state.flags &= ~SS_USE_SESSION_ID;
  }
  if (handle_PEP(in_buffer, hl) == -1) 
	send_reply(420, 0);
  remove_msg(hl + bl);
}

void
RTSP_Player::handle_play_reply(int status)
{
  int             hl;		/* message header length */
  int             bl;		/* message body length */
  long            l;

  get_msg_len(&hl, &bl);

  if (status < 203) {
    /*
     * if session ID was returned and does not match the ID returned on
     * previous setup response let the user know.  Otherwise, add the new
     * session ID now to the current settings.
     */
    if (get_int(in_buffer, hl, (char *) sSession, ": ", 0, &l)) {
      if (!(p_state.flags & SS_USE_SESSION_ID)) {
	p_state.flags |= SS_USE_SESSION_ID;
	p_state.settings.session_id = l;
      } else if ((long) p_state.settings.session_id != l)
	printf("Invalid Session ID returned by server.");
    }
    /*
     * here is where you would check for new sequence numbers for each
     * stream.  This reference implementation does not track starting
     * sequence numbers returned by the server.
     */
	if (handle_PEP(in_buffer, hl) == -1) { 
		remove_msg(hl + bl);	/* remove remainder of message from in_buffer */
		send_reply(420, "Bad Extension\n");
		return;
    }

    interface_start_play(p_state.settings.session_id);
  }
  remove_msg(hl + bl);
}

void
RTSP_Player::handle_pause_reply(int status)
{
  discard_msg();		/* remove remainder of message from in_buffer */
}

void
RTSP_Player::handle_close_reply(int status)
{
  discard_msg();		/* remove remainder of message from in_buffer */
  if (status < 203) {
    free_streams(&p_state);
    clear_udpfd();		/* close down udp socket connection. */
    interface_close_stream(p_state.settings.session_id);
    /* shut down the connection but leave application running. */
    Tcl_DeleteFileHandler(ssock_);
    tcp_close(ssock_);
    ssock_ = 0;
  }
}


/**********************************************************
 *    server socket connection and receive data handling
 *********************************************************/

int
RTSP_Player::server_connect(const char *server, u_short port, int *sock)
{
  int             got_host = 0;
  struct sockaddr_in addr;
  struct hostent *h = gethostbyname(server);
  if (h)
    got_host = 1;
  else {
    /* dot notation? */
    u_int32         net_addr = inet_addr(server);
    h = gethostbyaddr((const char *) &net_addr, sizeof(u_int32), PF_INET);
    if (h)
      got_host = 1;
  }
  if (got_host) {
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = *(u_int32 *) h->h_addr;
    addr.sin_port = htons(port);
    *sock = tcp_open((struct sockaddr *) & addr, sizeof(addr));
    if (*sock < 0)
      got_host = 0;
  }
  return got_host;
}


// read the frames from the playout buffer
void
RTSP_Player::read_adu() {

  u_int32_t addr;
  u_int16_t port;

  int dbuflen=0;
  char *dbuf=NULL;
  last_adu_++;
  int slot = (last_adu_) % ADU_BUFSIZE;

  if (adu_seqno[slot] != -1) {
#ifdef SRRTP_DEBUG_ENABLED
    fprintf(stderr, "reading frame %d (seq %d) for playout...\n",
	    last_adu_, adu_seqno[slot]);
#endif

    dbuf = (adubuf_[slot]->dbuf);
    dbuflen = adubuf_[slot]->dbuflen;

    // still must update left edge
    srrtp_update(rrtp_chan, last_adu_);


//      memcpy(dbuf, adubuf, adubuf->dbuflen);
//      dbuflen = adubuf->dbuflen;
//      free(adubuf);
//      adu_seqno[slot] = -1;
    
  } else {
#ifdef SRRTP_DEBUG_ENABLED
    fprintf(stderr, "forcing read of frame %d and cancelling further retransmits...\n", last_adu_);
#endif
    srrtp_read_and_cancel(rrtp_chan, last_adu_, (char *) dbuf,
			  &dbuflen, &addr, &port);
  }


  if (dbuf==NULL) {
#ifdef SRRTP_DEBUG_ENABLED
    fprintf(stderr, "frame read in %d ms\n", frametick_);
#endif
    if (last_adu_ < p_state.streams->duration)
      Tcl_CreateTimerHandler(frametick_, (Tcl_TimerProc*)adu_callback, this);
    
    return;
  }

//   MPEG-4 Processing code...do synchronously now!
  int code;
  memcpy (&code, dbuf, sizeof(int));
  code = htonl(code);

  if (show_bits_bytealign((unsigned char*)dbuf,32) == VOP_START_CODE) {
    decode_MPEG4_VOP(dbuf, dbuflen);

  } else if (!(code & 0x10)) {
    decode_MPEG4_VisualObjectStart(dbuf, dbuflen);
  } else 
    fprintf(stderr, "error: code 0x%x is not known\n", code);



//    free(dbuf);
  adu_seqno[slot]=-1;

#ifdef SRRTP_DEBUG_ENABLED
  fprintf(stderr, "frame read in %d ms\n", frametick_);
#endif
  if (last_adu_ < p_state.streams->duration)
    Tcl_CreateTimerHandler(frametick_, (Tcl_TimerProc*)adu_callback, this);
}


void
RTSP_Player::rtp_data_handle(int seqno, int len, int layer) {

  // difference here is that RTP channel has already received data from
  // the socket file descriptor so no need to make this call as in
  // udp_data_recv.  just reference data in packet appropriately.


  // note: some of the header info sets state, so we should probably also
  //       deal with that here
  
  ADU_Buffer *temp, *adubuf;

  u_int32_t addr;
  u_int16_t port;
  int dbuflen = len;
  int framerate = p_state.streams->framerate;
  
  char *dbuf = (char *) malloc(dbuflen);
  srrtp_read(rrtp_chan, seqno, (char *) dbuf, &dbuflen, &addr, &port);


  if (p_state.streams->mime_type != MIME_VIDEO)
    interface_stream_output(p_state.settings.session_id, dbuflen, dbuf);
  else {
#ifdef SRRTP_DEBUG_ENABLED
    printf("Process MPEG-4 Video ADU\n");
#endif
//      int code;
//      memcpy (&code, dbuf, sizeof(int));
//      code = htonl(code);
//      if (show_bits_bytealign((unsigned char*)dbuf,32) == VOP_START_CODE) {
//        decode_MPEG4_VOP(dbuf, dbuflen);
//      } else if (!(code & 0x10)) {
//        decode_MPEG4_VisualObjectStart(dbuf, dbuflen);
//      } else 
//        fprintf(stderr, "error: code 0x%x is not known\n", code);
    
    int slot = seqno%ADU_BUFSIZE;
    if (adu_seqno[slot] == -1) {
      temp = new ADU_Buffer(dbuf, dbuflen, seqno, layer);
      adubuf_[slot] = temp;
      adu_seqno[slot] = seqno;


      /* startup */
      if ((seqno > (framerate*ADU_BUF_MILLISECONDS/1000) &&
	  !playout_start_)) {
#ifdef SRRTP_DEBUG_ENABLED
	fprintf(stderr, "starting playout buffer timer...\n");
#endif
	playout_start_ = 1;
  	read_adu();
      }
      
    } else {
      fprintf(stderr, "WARNING: slots full, increase buffer size!\n");
    }
  }

//    free(dbuf);
}



//
// udp_data_recv:
//
// this is legacy stuff.  the old code used a callback waiting on the data socket
// and made the callback at the application layer, but this is wrong since
// this receive should be handled at the RRTP layer.
// INSTEAD:
//     using rtp_app_notify with rtp_data_handle callback


void
RTSP_Player::udp_data_recv(RTSP_SOCK fd)
{
  char            buffer[16384];
  int             n, l;
  RTP_HDR        *rtp;
  RTP_OP         *op;
  STREAM         *s;
  short           data_len;
  char           *data;
  int             addlen=0;
  u_short         seq_no;
  u_short         sflags;
  u_int32         oflags;

  n = recv(fd, buffer, sizeof(buffer), 0);
  if (n < 0)
    return;

  /* ignore data if it is not long enough to contain at least the RTP header */
  if (n < (int) sizeof(RTP_HDR)) {
    printf("ALERT: UDP data received did not include RTP header.\n");
    return;
  }
  rtp = (RTP_HDR *) buffer;

  if (rtp->payload != RTP_PAYLOAD_RTSP &&
      rtp->payload != RTP_PAYLOAD_L16_2 &&
      rtp->payload != RTP_PAYLOAD_L16_1 &&
	  rtp->payload != RTP_PAYLOAD_PCMA &&
      rtp->payload != RTP_PAYLOAD_PCMU) {
    printf("ALERT: Unsupported payload type: %d.\n",
	   rtp->payload);
    return;
  }
  /* reject if specific fields are not correct. */
  assert(p_state.streams);
  /*
   * stream id is expected in the SSRC field.  Match it to what we are
   * expecting.
   */
  for (s = p_state.streams; s && (s->stream_id != rtp->ssrc); s = s->next);
  if (!s) {
    printf("ALERT: RTP header rejected because SSRC not equal to stream id.\n");
    return;
  }
  seq_no = ntohs(rtp->seq_no);
  /*
   * This is where you could add code to reject the packet if the sequence
   * number is not correct.  This reference implementation does not implement
   * this mechanism.  The consequence is that any existing play range will
   * have to complete before the data for another play range will be heard.
   */
  s->seq_num = seq_no;		/* set this streams new sequence number */

  /* add code here to use time stamp. */

  if (rtp->csrc_len != 0) {
    printf("ALERT: RTP header rejected because CSRC length not 0.\n");
    return;
  }
  if (rtp->extension == 1) {
    addlen = sizeof(RTP_OP);
    op = (RTP_OP *) (buffer + sizeof(RTP_HDR));
    l = ntohs(op->op_code_data_length);
    if (ntohs(op->op_code) != RTP_OP_PACKETFLAGS)
      printf("ALERT: RTP extension header op_code invalid.\n");
    if (l == RTP_OP_CODE_DATA_LENGTH) {
      sflags = s->stream_flags;
      oflags = ntohl(op->op_code_data[0]);
      if (oflags & RTP_FLAG_LASTPACKET) {
	sflags |= STREAM_LAST_PACKET;
	/* if last packet and still in PLAY state return to READY state */
	if (p_state.cur_state == PLAY_STATE) {
	  p_state.cur_state = READY_STATE;
	  interface_stream_done(p_state.settings.session_id);
	}
      } else
	sflags &= ~STREAM_LAST_PACKET;

      if (oflags & RTP_FLAG_KEYFRAME)
	sflags |= STREAM_KEY_FRAME;
      else
	sflags &= ~STREAM_KEY_FRAME;
      s->stream_flags = sflags;
    } else
      printf("ALERT: RTP extension header op_code_data_length invalid.\n");
    addlen += (l * 4) - 4;
  } else {
    addlen = 0;
  }

  data = buffer + sizeof(RTP_HDR) + addlen;
  data_len = n - (sizeof(RTP_HDR) + addlen);

  if (debug_toggle) {
    printf("RTP_HEADER: ");
    dump_buffer(buffer, sizeof(RTP_HDR) + addlen + 2);
  }
  if (data_len) {
    if (PCM_hdr.wBitsPerSample == 16 && !big_endian()) {
      swap_word((unsigned short *) data, (data_len / 2));
    }
    interface_stream_output(p_state.settings.session_id, data_len, data);
  }
  return;
}


/***************************************
 *   player command handling routines
 **************************************/

void
RTSP_Player::handle_command(const char *c)
{
  char            cmd[256];
  char           *tok;
  char            ctok;
  static int      seen_open = 0;
  char            msg[512];
  char            s[2048];

#ifdef SRRTP_DEBUG_ENABLED
  printf("\nhandle_command: %s\n", c);
#endif
  strncpy(cmd, c, sizeof(cmd));
  tok = strtok(cmd, " \t\n");
  if (!tok) {
    interface_error("no command found in input.\n"
		    "Enter command ('?' or 'h' for help):\n");
    return;
  }
  ctok = tolower(*tok);
  if (ctok == 'o') {		/* open */
    char            url[1024];
    char            file_name[256];
    char            server_name[256];
    u_short         port;
    RTSP_SOCK       sock;
	/* BEGIN new variables added by Ed Hogan TIS Inc. */
      char file_name2[256];
      char server_name2[256];
      u_short port2;
      int using_a_proxy = 0;
    /* END   new variables added by Ed Hogan TIS Inc. */

    /* don't recurse */
    if (seen_open) {
      interface_error("\nalready connected\n");
      return;
    }
    tok = strtok(NULL, " \t\n");
    if (!tok) {
      interface_error("\nOPEN requires a URL in the form "
		      "'rtsp://server[:port][/file]'\n");
      return;
    }
    strcpy(url, tok);
    if (parse_url(url, server_name, &port, file_name)) {
		/* BEGIN Added By Ed Hogan, Trusted Information Systems Inc. */
         if (strlen(file_name) > 0) 
         {
            if (parse_url(file_name, server_name2, &port2, file_name2) &&
               (strlen(server_name2) != 0)) 
            {
               strcpy(p_state.serverurl, "rtsp://");
               strcat(p_state.serverurl, server_name2);
               strcat(p_state.serverurl, "/"); 
               /* port2 and file_name2 are discarded. sorry. */
               using_a_proxy = 1;
            }
         }
         if (!using_a_proxy) 
         {
	    strcpy(p_state.serverurl, "rtsp://");
	    strcat(p_state.serverurl, server_name);
            strcat(p_state.serverurl, "/");
         } 
         if(server_connect(server_name, port, &ssock_))
         {

	   Tcl_CreateFileHandler(ssock_, TCL_READABLE,
				 (Tcl_FileProc*)tcpread_callback, (ClientData)this); 
            if (using_a_proxy)
               handle_event( CMD_OPEN, 0, file_name );
            else /* Not using a proxy */
               handle_event( CMD_OPEN, 0, 0);
         /* END   Added By Ed Hogan, Trusted Information Systems Inc. */

	    strcpy(OpenFileName, file_name);
	    seen_open = 1;
	    //ssock_ = sock;
      } else {
	sprintf(msg, "\nunable to connect to server //%s:%d\n",
		server_name, port);
	interface_error(msg);
      }
    } else
      interface_error("\nunable to parse url\n");
  } else if (ctok == 'g') {	/* get */
    char            url[1024];
    char            file_name[256];
    char            server_name[256];
    u_short         port;

    tok = strtok(NULL, " \t\n");
    if (!tok) {
      sprintf(msg, "\nGET requires an URL in the form "
	      "'path/mediafile.wav'\n");
      interface_error(msg);
      return;
    }
    strcpy(url, tok);
    if (parse_url(url, server_name, &port, file_name))
      handle_event(CMD_GET, 0, file_name);
    else
      interface_error("\nunable to parse url\n");
  } else if (ctok == 'p') {	/* play */
    char            range[64];

    tok = strtok(NULL, " \t\n");
    *range = '\0';
    if (tok) {
      if (strchr(tok, '=')) {
	strncpy(range, tok, 63);
	range[63] = '\0';
      } else {			/* range doesn't have smpte prefix */
	strcpy(range, "smpte=");
	strcat(range, tok);
      }
    }
    handle_event(CMD_PLAY, 0, range);
  } else if (ctok == 'z') {	/* pause */
    handle_event(CMD_PAUSE, 0, 0);
  } else if (ctok == 'c') {	/* close */
    handle_event(CMD_CLOSE, 0, 0);
    seen_open=0;
  } else if (ctok == 'd') {	/* debug toggle */
    debug_toggle = debug_toggle ? 0 : 1;
    sprintf(s, "Debug toggle is %s.\n", debug_toggle ? "ON" : "OFF");
    interface_alert_msg(s);
  } else if (ctok == 'q' || ctok == 'x') {	/* quit, exit */
    Quit = 1;			/* this will disable the timer event handler */
    exit(0);
  } else if (ctok == 'h' || ctok == '?') {
    sprintf(s, "Player commands:\n");
    sprintf(s + strlen(s), "                 format & examples of usage\n");
    sprintf(s + strlen(s), " o)pen     establishes RTSP control channel with the specified server.\n"
	    "     format:     o [rtsp://]server[:port]\n"
	    "     examples:   o bigserver\n"
	    "                 o rtsp://mediaserver:9095\n");
    sprintf(s + strlen(s), " g)et      gets the specified file's description and\n"
	    "           sets up server and player to play.\n"
	    "     format:     g [[/]path/]filename\n"
	    "     examples:   g audio.wav\n"
	    "                 g /sounds_dir/drip.wav\n"
	    "                 g ascii.txt\n");
    sprintf(s + strlen(s), " p)lay     retrieves and plays streamed audio data. A\n"
	    "           range may be specified in smpte format.\n"
	    "     format:     p [[smpte=]00:00:10-[00:01:30]]\n"
	    "     examples:   p\n"
	    "                 p 00:00:20-\n"
	    "                 p smpte=00:00:20-00:01:30\n");
    sprintf(s + strlen(s), " z) pause        sends pause request to server.\n"
	    "     example:    z\n");
    sprintf(s + strlen(s), " c)lose          c\n");
    sprintf(s + strlen(s), " d)ebug toggle   d\n");
    sprintf(s + strlen(s), " h)elp           h\n");
    sprintf(s + strlen(s), " q)uit           q\n"
	    "                 x\n");
    interface_alert_msg(s);
  } else {
    interface_error("\nunrecognized command\n");
  }
}


/***************************************
 *       player state machine
 **************************************/

void
RTSP_Player::handle_event(int event, int status, char *buf)
{
  static const char sStateErrMsg[] = "State error: event %d in state %d\n";
  char            msg[2048];

  /*
   * All state transitions are made here except when the last stream packet
   * is received during a PLAY.  That transition is located in
   * udp_data_recv().
   */
  switch (p_state.cur_state) {
  case INIT_STATE:
    {
      switch (event) {
      case RTSP_GET_RESPONSE:
	if (handle_get_reply(status))
	  send_setup_request();	/* now send a setup request */
	break;

      case RTSP_SETUP_RESPONSE:
	handle_setup_reply(status);
	if (status < 203) {
	  p_state.cur_state = READY_STATE;
	  handle_event(CMD_PLAY, 0, 0);
	}
	break;

      case RTSP_CLOSE_RESPONSE:
	handle_close_reply(status);
	break;

      case RTSP_HELLO_RESPONSE:
	handle_hello_reply(status);
	break;

      case RTSP_HELLO_METHOD:
	handle_hello_request();
	break;

      case CMD_OPEN:
	send_hello_request(buf);
	break;

      case CMD_GET:
	send_get_request(buf);
	break;

      case CMD_CLOSE:
	send_close_request();
	break;

      case RTSP_SESSION_METHOD:/* method not valid this state. */
	send_reply(455, "Accept: HELLO\n");
	break;

      default:
	sprintf(msg, sStateErrMsg, event, p_state.cur_state);
	interface_error(msg);
	/*
	 * if this event is not from user input then remove the message from
	 * the receive buffer.
	 */
	if (event < CMD_OPEN) {
	  discard_msg();	/* remove remainder of message from in_buffer */
	  if (event < RTSP_SETUP_RESPONSE)	/* method not implemented */
	    send_reply(501, "Accept: HELLO\n");
	}
	break;
      }
      break;
    }				/* INIT state */

  case READY_STATE:
    {
      switch (event) {
      case RTSP_PLAY_RESPONSE:
	handle_play_reply(status);
	if (status < 203)
	  p_state.cur_state = PLAY_STATE;
	break;

      case RTSP_GET_RESPONSE:
	if (handle_get_reply(status)) 
	  send_setup_request();	
	else 
		p_state.cur_state = INIT_STATE;
	break;
		
	  case RTSP_SETUP_RESPONSE:
	handle_setup_reply(status);
	if (status < 203) 
	  handle_event(CMD_PLAY, 0, 0);
	else
		 p_state.cur_state = INIT_STATE;
	break;

      case RTSP_CLOSE_RESPONSE:
	handle_close_reply(status);
	if (status < 203)
	  p_state.cur_state = INIT_STATE;
	break;

      case RTSP_HELLO_RESPONSE:
	handle_hello_reply(status);
	break;

      case RTSP_HELLO_METHOD:
	handle_hello_request();
	break;

      case CMD_PLAY:
	send_play_range_request(buf);
	break;

      case CMD_GET:		/* allow get w/o changing state. */
	send_get_request(buf);
	break;

      case CMD_CLOSE:
	send_close_request();
	break;

      default:
	sprintf(msg, sStateErrMsg, event, p_state.cur_state);
	interface_error(msg);
	if (event < CMD_OPEN) {
	  discard_msg();	/* remove remainder of message from in_buffer */
	  if (event < RTSP_SETUP_RESPONSE)	/* method not implemented */
	    send_reply(501, "Accept: HELLO\n");
	}
	break;
      }
      break;
    }				/* READY state */

  case PLAY_STATE:
    {
      switch (event) {
      case RTSP_PLAY_RESPONSE:
	handle_play_reply(status);
	if (status >= 203)
	  p_state.cur_state = READY_STATE;
	break;

      case RTSP_PAUSE_RESPONSE:
	handle_pause_reply(status);
	if (status < 203)
	  p_state.cur_state = READY_STATE;
	break;

      case RTSP_CLOSE_RESPONSE:
	handle_close_reply(status);
	if (status < 203)
	  p_state.cur_state = INIT_STATE;
	break;

      case RTSP_HELLO_RESPONSE:
	handle_hello_reply(status);
	break;

	case RTSP_GET_RESPONSE:
	if (handle_get_reply(status)) 
	  send_setup_request();	
	else 
		p_state.cur_state = INIT_STATE;
	break;
		
	  case RTSP_SETUP_RESPONSE:
	handle_setup_reply(status);
	if (status < 203) 
	  handle_event(CMD_PLAY, 0, 0);
	else
		 p_state.cur_state = INIT_STATE;
	break;

      case RTSP_HELLO_METHOD:
	handle_hello_request();
	break;

      case CMD_PLAY:
	send_play_range_request(buf);
	break;

      case CMD_PAUSE:
	send_pause_request();
	break;

      case CMD_GET:		/* allow get w/o changing state. */
	// Added by NGF 8/2/00 -- upon new get request, want to reset wav settings, etc.
	if (wav_fd)
	  interface_stream_done(p_state.settings.session_id);
	send_get_request(buf);
	break;

      case CMD_CLOSE:
	send_close_request();
	break;

      default:
	sprintf(msg, sStateErrMsg, event, p_state.cur_state);
	interface_error(msg);
	if (event < CMD_OPEN) {
	  discard_msg();	/* remove remainder of message from in_buffer */
	  if (event < RTSP_SETUP_RESPONSE)	/* method not implemented */
	    send_reply(501, "Accept: HELLO\n");
	}
	break;
      }
      break;
    }				/* PLAY state */

  default:			/* invalid/unexpected current state. */
    {
      sprintf(msg, "State error: unknown state=%d, event code=%d\n",
	      p_state.cur_state, event);
      interface_error(msg);
      if (event < CMD_OPEN)
	discard_msg();		/* remove remainder of message from in_buffer */
    }
    break;
  }				/* end of current state switch */
}





/******************************** User Interface Methods ******************************/

#ifdef TCL_GUI
int
url_cb(ClientData *cd, Tcl_Interp *interp, int argc, char **argv)
{
    char buf[1024];
    RTSP_Player *thisp = (RTSP_Player*)cd;

    strcpy(thisp->url, argv[1]);

    if (thisp->hello_yet)
    {
      sprintf(buf, "get %s\n", thisp->url);
      thisp->handle_command(buf);
    }
    else 
    {
      sprintf(buf, "open %s\n", thisp->url);
      thisp->handle_command(buf);
      sprintf(buf, "get %s\n", thisp->url);
      thisp->handle_command(buf);
    }
    return TCL_OK;
}


int
play_cb(ClientData *cd, Tcl_Interp *interp, int argc, char **argv)
{
    char buf[1024];
    RTSP_Player *thisp = (RTSP_Player*)cd;

    sprintf(buf, "play %s", argv[1]);
    thisp->handle_command(buf);
    if (atoi(argv[2]))
    {
    	if (!thisp->wav_session)
	    thisp->wav_session = atoi(argv[1]);
    	else
    	{
    	    sprintf(buf, ".ns%d.debug insert end \"Audio device in use, using streaming text mode\n\"", atoi(argv[1]));
    	    Tcl_Eval(interp, buf);
    	}
    }

    return TCL_OK;
}


int
stop_cb(ClientData *cd, Tcl_Interp *interp, int argc, char **argv)
{
    char buf[1024];
    RTSP_Player *thisp = (RTSP_Player*)cd;

    sprintf(buf, "close %s", argv[1]);
    thisp->handle_command(buf);
    thisp->hello_yet=0;
    return TCL_OK;
}


int
resume_cb(ClientData *cd, Tcl_Interp *interp, int argc, char **argv)
{
    char buf[1024];
    RTSP_Player *thisp = (RTSP_Player*)cd;

    sprintf(buf, "play %s", argv[1]);
    thisp->handle_command(buf);

    return TCL_OK;
}

int
speed_cb(ClientData *cd, Tcl_Interp *interp, int argc, char **argv)
{
#if defined AUDIO
    char buf[1024];
    int x;
    RTSP_Player *thisp = (RTSP_Player*)cd;


    x = 65536 + (atoi(argv[2]) - 500) * 130;

    sprintf(buf, "z %s %d", argv[1], x);

    // NGF -- HACK
    //  thisp->handle_command(buf);

    if (thisp->wav_session == atoi(argv[1]))
    {
    	x = (int)(thisp->wav_rate * (x / 65536.0));
	ioctl(thisp->wav_fd, SNDCTL_DSP_SPEED, &x);
    }
#endif
    return TCL_OK;
}

int
quit_cb(ClientData *cd, Tcl_Interp *interp, int argc, char **argv)
{
  RTSP_Player *thisp = (RTSP_Player*)cd;
  thisp->handle_command("quit\n");

  return TCL_OK;
}
#endif // TCL_GUI




void
RTSP_Player::interface_init()
{
#ifdef TCL_GUI
  interp = Tcl_CreateInterp();
  
  if (Tcl_Init(interp) == TCL_ERROR)
    {
      fprintf (stderr, "Fatal: Tcl_Init Failed\n");
      exit (1);
    }
  
#ifdef TK40
  Tk_CreateMainWindow(interp, 0, "rtspref", "Rtspref");
#endif
  
  if (Tk_Init(interp) == TCL_ERROR)
    {
      fprintf (stderr, "Fatal: Tk_Init Failed\nCheck Your DISPLAY\n");
      exit (1);
    }
  
  Tcl_EvalFile(interp, "gui.tcl");
  Tcl_CreateCommand(interp, (char *)"url_cb", (Tcl_CmdProc *)url_cb, (ClientData)this, 0);
  Tcl_CreateCommand(interp, (char *)"Play", (Tcl_CmdProc *)play_cb, (ClientData)this, 0);
  Tcl_CreateCommand(interp, (char *)"Stop", (Tcl_CmdProc *)stop_cb, (ClientData)this, 0);
  Tcl_CreateCommand(interp, (char *)"Resume", (Tcl_CmdProc *)resume_cb, (ClientData)this, 0);
  Tcl_CreateCommand(interp, (char *)"Speed", (Tcl_CmdProc *)speed_cb, (ClientData)this, 0);
  Tcl_CreateCommand(interp, (char *)"Quit", (Tcl_CmdProc *)quit_cb, (ClientData)this, 0);
  //    Tk_CreateTimerHandler(10, (Tk_TimerProc *)event, 0);
#endif // TCL_GUI
}

void 
RTSP_Player::interface_can_play(int id)
{
#ifndef TCL_GUI
    printf("Can play %d\n", id);
#endif
}

void 
RTSP_Player::interface_close_stream(int id)
{
#ifndef TCL_GUI
    printf( "CLOSE\n" );
#endif
}

void 
RTSP_Player::interface_new_session(int id)
{
#ifdef TCL_GUI
    char buf[1024];

    sprintf(buf, "NewSession %ld", id);
    // open the session window
    //Tcl_Eval(interp, buf);
#else
    printf("New session: %d\n", id);
#endif
}

void 
RTSP_Player::interface_close_session(int id)
{
#ifdef TCL_GUI
  char buf[1024];

  sprintf(buf, "destroy .ns%ld", id);
  Tcl_Eval(interp, buf);
  
  close (wav_fd);
  wav_fd = 0;
  wav_session = 0;
#else
  printf("Close session: %d\n", id);
#endif
}

void 
RTSP_Player::interface_hello(const char *filename)
{
#ifdef TCL_GUI
  char buf[1024];
  
  hello_yet++;
  sprintf(buf, ".debug insert end \"Got HELLO\n\"\n");
  Tcl_Eval(interp, buf);
  
  if (url[0])
    {
      sprintf(buf, "open %s\n", url);
      handle_command(buf);
    }
#else
  printf ("HELLO\n");
  if (strlen(filename)) {
    char cmdbuf[20];
    sprintf(cmdbuf, "g %s", filename);
    handle_command(cmdbuf);
  }
#endif
}

void 
RTSP_Player::interface_renew_play( u_long id )
{
#ifndef TCL_GUI
   printf( "Enter new command\n" );
#endif
}

void 
RTSP_Player::interface_start_play( u_long id )
{
#ifndef TCL_GUI
   printf( "PLAY started.\n" );
#endif
}

#ifdef AUDIO
void
convert16to8(u_short n, char *data)
{
  int i;

  for (i=0;i<=(n/2);i++) {
    data[i]=data[i*2]-128;
  }
}

void
RTSP_Player::interface_stream_done( u_long sessionId )
{
#ifndef TCL_GUI
    printf( "PLAY done.\n" );
#endif
    
#ifdef AUDIO
    close (wav_fd);
    wav_fd = 0;
    wav_session = 0;
#endif  /* AUDIO */
}

void
RTSP_Player::interface_error( const char *err )
{
#ifndef TCL_GUI
    printf( err );
#endif
}

void
RTSP_Player::interface_alert_msg( const char *msg )
{
#ifndef TCL_GUI
    printf( msg );
#endif
}


#ifndef TCL_GUI
void
RTSP_Player::interface_stream_output(u_long id, u_short n, char *data)
{
    if (wav_session == id) {
    	if (!wav_fd)
    	{
    	    int x;

    	    wav_fd = open("/dev/dsp", O_WRONLY);
    	    wav_rate = PCM_hdr.nSamplesPerSec;
            if(PCM_hdr.wBitsPerSample==16) {
              x = AFMT_S16_BE;
	      if(ioctl(wav_fd, SNDCTL_DSP_SETFMT, &x)==-1)
              {
                 printf("Error: Cannot set sample format\n");
		 exit(1);
              }
	      wav_sampleformat = x; 
              printf("Sample Format (supported by card): %d\n", x);
	    }
	    else if (PCM_hdr.wFormatTag==0x7)
	    {
	      /*  muLaw  */
	      x = AFMT_MU_LAW;
	      if(ioctl(wav_fd, SNDCTL_DSP_SETFMT, &x)==-1)
              {
                 printf("Error: Cannot set sample format\n");
		 exit(1);
              }
	      wav_sampleformat = x; 
	    }
	    else
	    {
	      /*  assume unsigned 8-bit, and that /dev/dsp is all setup */
	      wav_sampleformat = AFMT_U8;
	    }
	      
    	    x = PCM_hdr.nChannels -1;	
	    if (ioctl(wav_fd, SNDCTL_DSP_STEREO, &x)==-1) {
                 printf("Error: Cannot set channels\n");
		 //exit(1);  
	    }
    	    x = wav_rate;		
	    if (ioctl(wav_fd, SNDCTL_DSP_SPEED, &x)==-1) {
                 printf("Error: Cannot set speed\n");
	    }

	    printf("WAV File\n");
	    printf("Frequency: %ld\n", PCM_hdr.nSamplesPerSec);
	    printf("Channels: %d\n", PCM_hdr.nChannels);
            printf("Bits Per Sample: %d\n", PCM_hdr.wBitsPerSample);
    	}
	
	if(PCM_hdr.wBitsPerSample==16 && wav_sampleformat == AFMT_U8 ) 
        {
	  convert16to8(n, data);
	  write (wav_fd, data, (n/2));
	}
	else
        {
	  write (wav_fd, data, n);
	}
    	return;
    }
}
#endif /* AUDIO */

#ifndef AUDIO
void
RTSP_Player::interface_stream_output(u_long id, u_short n, const char *data)
{
   static int  created = 0;
   int         fd;
   char        filename[256];
   int         flags;
   int         cnt;
   
   /* append output to file named 'audioXXXX.txt' */
   sprintf(filename, "audio%ld.dat", id);
   if ( created )
      flags = O_WRONLY | O_APPEND;
   else
   {
      flags = O_CREAT | O_WRONLY | O_TRUNC;
      created = 1;
   }

   fd = open( filename, flags, 0x1a4 );  /* rw/r/r permissions */
   if ( fd == -1 )
   {
      printf( "ALERT:  error opening file interface_stream_output(). "
              "errno=%d", errno );
      return;
   }
   cnt = write( fd, data, n );
   if ( cnt == -1 )
      printf( "ALERT: error writing interface_stream_output(). "
              "errno=%d", errno );
   else if ( cnt != n )
      printf( " Error write %d byte(s) to %s. Only wrote %d byte(s).",
              n, filename, cnt );
   close( fd );
}
#endif  /* ifndef AUDIO */
#endif /* ifndef TCL_GUI */


#ifdef TCL_GUI


void
RTSP_Player::interface_stream_output(u_long id, u_short n, char *data)
{
    char buf[1024];

#ifdef AUDIO
    /* XXX - Packet size must be large enough to transmit the whole header */
    /* XXX - We could really use some buffering :-) */

    if (wav_session == id) {
    	if (!wav_fd)
    	{
    	    int x;

    	    wav_fd = open("/dev/dsp", O_WRONLY);
    	    wav_rate = PCM_hdr.nSamplesPerSec;
            if(PCM_hdr.wBitsPerSample==16) {
              x = AFMT_S16_BE;
	      if(ioctl(wav_fd, SNDCTL_DSP_SETFMT, &x)==-1)
              {
                 printf("Error: Cannot set sample format\n");
		 exit(1);
              }
	      wav_sampleformat = x; 
              printf("Sample Format (supported by card): %d\n", x);
	    }
	    else if (PCM_hdr.wFormatTag==0x7)
	    {
	      /*  muLaw  */
	      x = AFMT_MU_LAW;
	      if(ioctl(wav_fd, SNDCTL_DSP_SETFMT, &x)==-1)
              {
                 printf("Error: Cannot set sample format\n");
		 exit(1);
              }
	      wav_sampleformat = x; 
	    }
	    else
	    {
	      /*  assume unsigned 8-bit, and that /dev/dsp is all setup */
	      wav_sampleformat = AFMT_U8;
	    }
	      
    	    x = PCM_hdr.nChannels -1;	
	    if (ioctl(wav_fd, SNDCTL_DSP_STEREO, &x)==-1) {
                 printf("Error: Cannot set channels\n");
		 //exit(1);
	    }
    	    x = wav_rate;		
	    if (ioctl(wav_fd, SNDCTL_DSP_SPEED, &x)==-1) {
                 printf("Error: Cannot set speed\n");
	    }

	    sprintf(buf, ".ns%ld.output insert end \"WAV File\n\"", id);
    	    Tcl_Eval(interp, buf);
	    sprintf(buf, ".ns%ld.output insert end \"Frequency: %ld\n\"", 
		    id, PCM_hdr.nSamplesPerSec);
    	    Tcl_Eval(interp, buf);
	    sprintf(buf, ".ns%ld.output insert end \"Channels: %d\n\"", 
		    id, PCM_hdr.nChannels);
    	    Tcl_Eval(interp, buf);
	    sprintf(buf, ".ns%ld.output insert end \"Bits per sample: %d\n\"", 
		    id, PCM_hdr.wBitsPerSample);
    	    Tcl_Eval(interp, buf);
    	}
	if(PCM_hdr.wBitsPerSample==16 && wav_sampleformat == AFMT_U8 ) 
        {
	  convert16to8(n, data);
	  write (wav_fd, data, (n/2));
	}
	else
        {
	  write (wav_fd, data, n);
	}
    	return;
    }
#endif
#ifndef AUDIO
    u_short i;

    sprintf(buf, ".ns%ld.output insert end \"", id);
    for (i = 0; i < n; i++)
    {
    	if (data[i] != '\n')
    	{
    	    int x = strlen(buf);
    	    buf[x] = data[i];
    	    buf[x+1] = 0;
    	}
    	else
    	{
    	    strcat(buf, "\n\"\n");
	    Tcl_Eval(interp, buf);
	    sprintf(buf, ".ns%ld.output insert end \"", id);
    	}
    }
    strcat(buf, "\"\n");
    Tcl_Eval(interp, buf);

    sprintf(buf, ".ns%ld.output see end\n", id);
    Tcl_Eval(interp, buf);
#endif
}


#endif /* TCL_GUI */
